// ClientConnection.cpp

#include "EQServerSocket_stdafx.h"

#include "ClientConnection.h"
#include "UDPServerSocket.h"
#include "EQTxPacket.h"
#include "hex.h"
#include "everquest.h"

CClientConnection::CClientConnection(CUDPServerSocket &serverSocket, sockaddr_in addr)
	: m_serverSocket(serverSocket)
	, m_addr(addr)

	, m_LastSentTick()
	, m_LastPacketRecieved()

	, m_NextRxSeqID(0)
	, m_NextRxFragID(0)

	, m_NextTxFragID(0)
	, m_NextTxSeqID(0)
	, m_NextTxPacketID(0)
	, m_NextTxMsgID(0)

	, m_NeedsSYN(true)

{
	m_RecvQueue.clear();
	m_SendQueue.clear();
}

CClientConnection::~CClientConnection()
{
	CSingleLock localLock(&m_lockObj, TRUE);

	while(m_SendQueue.size())
	{
		delete m_SendQueue[m_SendQueue.size() - 1];
		m_SendQueue.pop_back();
	}

	while(m_RecvQueue.size())
	{
		delete m_RecvQueue[m_RecvQueue.size() - 1];
		m_RecvQueue.pop_back();
	}

	while(m_OrderedRecvQueue.size())
	{
		delete m_OrderedRecvQueue[m_OrderedRecvQueue.size() - 1];
		m_OrderedRecvQueue.pop_back();
	}
}

// Every time a new packet is recieved that is seqenced, I iterate through
// the recieve queue to check for other packets that can be processed.
// If the new packet is fragmented, I check through those as well to
// see if I can assemble a full message.

// This prevents me from having to have a separate function that I call
// to perform these operations. Furthermore, I limit the number of times
// that I perform these operations to only times when it's applicable.
void CClientConnection::newPacket(CEQPacket &NewPacket)
{
	CSingleLock localLock(&m_lockObj, TRUE);

	// Check to see if we already got this msg? (duped)

	{
		CSingleLock localLock(&m_serverSocket.m_logFileSync, true);

		if(m_serverSocket.m_fDebugLog)
		{
			fprintf(m_serverSocket.m_fDebugLog, "Received %s:\n", (const char*)CXTime().toString());
			hexdump(NewPacket.getData().getData(), NewPacket.getData().getDataSize(), 16, m_serverSocket.m_fDebugLog);
			fflush(m_serverSocket.m_fDebugLog);
		}
	}

	int FragPartCount;
	static CEQPacket *fragments[MAX_FRAGMENTS];
	int i;
	size_t totalSize;
	bool packetUsed = false;

	// Disassemble the packet and discard if corrupt or malformed.
	if(!NewPacket.DisAssemblePacket())
		return;

	ClearTimeOut();

	// If fragmented, setup for frag packet...
	if(NewPacket.isFragmented())
	{
		// too many fragments maybe DoS attack or corrupt packet.
		if(NewPacket.getMaxFragIndex() >= MAX_FRAGMENTS)
			return;

		FragPartCount = 1;
		totalSize = NewPacket.getMsgDataLen();
		fragments[NewPacket.getFragIndex()] = &NewPacket;
	}

	// If the SYN flag is set then I need to sync up with the client (maybe
	// client is just connecting)... Also, we should SYN back.
	if(NewPacket.isSyncronize())
	{
		m_NextRxSeqID = NewPacket.getSequenceID();
	}

	// Check if this is packet is ACKing something we sent...
	if(NewPacket.isAcknowledge())
	{
		// remove the packet that is getting ACKed by the client.
		for(int i = 0; i < m_SendQueue.size(); i++)
		{
			if(m_SendQueue[i]->getSequenceID() == NewPacket.getAcknowledgeID())
			{
				delete m_SendQueue[i];
				m_SendQueue.remove(i);
				break;
			}
		}
		// oops, I didn't find it... oh well....
	}

	if(NewPacket.isFinish())
	{
		// Ok, Love ya, bye
		m_serverSocket.RemoveClientSocket(this);
	}

	// If packet is sequenced then we need to send the client an ACK.
	if(NewPacket.isSequenced())
	{
		// Add the seq id to the ACK queue.
		m_AckQueue.push_back(CAck(NewPacket.getSequenceID()));
	}
	else
	{	// If the packet isn't SEQ then quickly toss it into the ordered
		// recieve queue (if it has a high level message) and exit stage left.
		if(NewPacket.isMessage())
			MessageRecieved(new CBinary(NewPacket.getMsgDataLen(), NewPacket.getMsgData()));

		// done with packet...
		return;
	}

	// Quick check to catch most sequenced packets (because they will
	// usually be in order). Can't do this for fragmenteded tho.
	if(	NewPacket.isSequenced() &&
		NewPacket.getSequenceID() == m_NextRxSeqID &&
		!NewPacket.isFragmented())
	{
		// Add the seq non-frag packet into the ordered recieve queue, inc
		// m_NextRxSeqID and exit.
		if(NewPacket.isMessage())
			MessageRecieved(new CBinary(NewPacket.getMsgDataLen(), NewPacket.getMsgData()));
		m_NextRxSeqID++;
		packetUsed = true;
	}

	// detect dupe.
	if(NewPacket.isSequenced() && NewPacket.getSequenceID() < m_NextRxSeqID)
	{
		// we already got this one and we have just acked it so that's it....
		return;
	}

	// If a SEQ packet that wasn't fragged makes it this far then it'll end
	// up queued. All fraged packets will enter this loop.

	// Iterate through the recieve queue. Process any old packets that I can
	// and deal with fragmented packets.
	for (i = 0; i < m_RecvQueue.size(); i++)
	{
		// Check for a SEQ packet that isn't fragmented that was recieved out
		// of order, but who's predecessor has been recieved now.
		if(	m_RecvQueue[i]->isSequenced() &&
			m_RecvQueue[i]->getSequenceID() == m_NextRxSeqID &&
			!m_RecvQueue[i]->isFragmented())
		{
			if(NewPacket.isMessage())
				MessageRecieved(new CBinary(m_RecvQueue[i]->getMsgDataLen(), m_RecvQueue[i]->getMsgData()));

			m_NextRxSeqID++;
			delete m_RecvQueue[i];
			m_RecvQueue.remove(i);
			i--;
			continue;
		}

		// SEQed packets are taken care of now. Time to work on fragmented.
		if(	!packetUsed &&
			NewPacket.isFragmented() &&
			m_RecvQueue[i]->isFragmented() &&
			m_RecvQueue[i]->getFragID() == NewPacket.getFragID())
		{
			// check for dupe.
			if(!packetUsed && m_RecvQueue[i]->getFragIndex() == NewPacket.getFragIndex())
			{
				// oops, this isn't very usefull, toss it out

				return;
			}

			FragPartCount++;
			fragments[m_RecvQueue[i]->getFragIndex()] = m_RecvQueue[i];
			totalSize += m_RecvQueue[i]->getMsgDataLen();

			// Check to see if we got all the fragments.
			if(FragPartCount == NewPacket.getMaxFragIndex())
			{
				CBinary tmp(totalSize);
				tmp.setDataSize(0);
				UWORD maxSeqID = fragments[NewPacket.getMaxFragIndex() -1]->getSequenceID();

				// update the m_NextRxSeqID accordingly.
				if(m_NextRxSeqID <= maxSeqID)
					m_NextRxSeqID = maxSeqID + 1;

				// m_NextRxSeqID = NewPacket.getMaxFragIndex(); // ??

				for(int j = 0; j < NewPacket.getMaxFragIndex(); j++)
				{
					tmp.append(fragments[j]->getMsgData(), fragments[j]->getMsgDataLen());
					if(fragments[j] != &NewPacket)
					{
						delete fragments[j];
						m_RecvQueue.remove(m_RecvQueue.indexOf(fragments[j]));
						i--;
					}
				}
				// Ok, the message has been reassembled... now put it in the ordered queue.
				if(NewPacket.isMessage())
					MessageRecieved(new CBinary(tmp));

				packetUsed = true;
			}
		}
	}

	// If I didn't use the packet, stick it (actually a copy of it) in the recieve queue.
	if(!packetUsed)
	{
		// check for condition where I get a fragment that is in order and
		// need to increment the seq id.
		if(NewPacket.getSequenceID() == m_NextRxSeqID)
			m_NextRxSeqID++;

		m_RecvQueue.push_back(new CEQPacket(NewPacket));
	}
}

// Check to see if we need to resend packets that haven't been ACKed.
// Check to see if we haven't heard from Mr. Client in a while.
// Then check to see if we have outgoing ACKs that haven't been
// sent (that need to be sent soon)
void CClientConnection::DoTick()
{
	CSingleLock localLock(&m_lockObj, TRUE);

	// Check for timeout condition
	if(m_LastPacketRecieved.getMillisSince() > TIMEOUT * 1000)
	{
		// ok, he's LD
		m_serverSocket.RemoveClientSocket(this);

		// Send a FIN message.
		Send(new CEQTxPacket(
				EQ_MSG_FIN,
				m_NextTxPacketID++,
				0,
				this,
				0));
		return;
	}

	// Check for packets that weren't ACKed
	for(int i = 0; i < m_SendQueue.size(); i++)
		if(m_SendQueue[i]->getTimeSent().getMillisSince() >= RESEND_TIME)
		{
			m_SendQueue[i]->setPacketID(m_NextTxPacketID++);
			SendImpl(m_SendQueue[i]);
			m_SendQueue[i]->getTimeSent().reset();
		}

	// Send keep-alive ping if needed.
	if(m_LastSentTick.getMillisSince() >= KEEP_ALIVE_PING_INTERVAL)
	{
		CBinary tmp(1);
		tmp.setDataSize(0);
		//Send(&tmp, EQ_MSG_SEQ | EQ_MSG_SYS);
	}

	// Check for ACKs that need to go out.
	while(m_AckQueue.size() && m_AckQueue[0].getTimeReceived().getMillisSince() > ACK_QUEUE_TIME)
	{
		Send(new CEQTxPacket(
				EQ_MSG_ACK,
				m_NextTxPacketID++,
				m_AckQueue[0].getAckID(),
				this,
				0));
		m_AckQueue.remove(0);
	};

}

bool CClientConnection::IsPacketQueued()
{
	CSingleLock localLock(&m_lockObj, TRUE);

	return m_OrderedRecvQueue.size() ? true : false;
}

CBinary *CClientConnection::GetRecvPacket()
{
	CSingleLock localLock(&m_lockObj, TRUE);

	if(m_OrderedRecvQueue.size())
	{
		CBinary *ret = m_OrderedRecvQueue[0];
		m_OrderedRecvQueue.remove(0);
		return ret;
	}

	return NULL;
}

// Send a CBinary to the client. Will split large messages into fragments
// and handle seqencing. Also sends out ACKs from ACK queue by attaching
// them to the outgoing packets.
void CClientConnection::Send(CBinary *txMsg, unsigned short headers)
{
	try
	{
		CSingleLock localLock(&m_lockObj, TRUE);

		unsigned short ackID = 0;

		// if I'm expecting a response from this, I can use this instead of
		// heartbeat
		if(headers & EQ_MSG_SEQ)
			m_LastSentTick.reset();

		if(this->m_AckQueue.size())
		{
			headers |= EQ_MSG_ACK;
			ackID = m_AckQueue[0].getAckID();
			m_AckQueue.remove(0);
		}
		else
			headers &= ~EQ_MSG_ACK;

		// If this is the 1st SEQ packet, then we need to sent SYN to
		// sync the client.
		if(m_NeedsSYN && (headers & EQ_MSG_SEQ))
		{
			headers |= EQ_MSG_SYN;
			m_NeedsSYN = false;
		}

		if(txMsg->getDataSize() < MAX_PACKET_SIZE)
		{
			if(headers & EQ_MSG_SEQ)
			{
				Send(new CEQTxPacket(
					headers,
					m_NextTxPacketID++,
					ackID,
					m_NextTxSeqID++,
					0,
					0,
					0,
					(headers & EQ_MSG_MSG ? m_NextTxMsgID++ : 0),
					txMsg->getData(),
					txMsg->getDataSize()));
			}
			else
			{
				Send(new CEQTxPacket(
					headers,
					m_NextTxPacketID++,
					ackID,
					txMsg->getData(),
					txMsg->getDataSize()));
			}
		}
		else
		{	// Fragment the message and send the fragments...

			bool isMsg = headers & EQ_MSG_MSG ? true : false;

			unsigned int fragIndex;
			unsigned int numFrags = (txMsg->getDataSize() + (isMsg ? 2 : 0)) / MAX_PACKET_SIZE + 1;
			char *fragStart;
			unsigned int fragSize;

			for(fragIndex = 0; 
				fragIndex < numFrags;
				fragIndex++)
			{
				// If this is not the 1st fragment...
				if(fragIndex != 0)
				{
					// Recheck ACK queue and ACK another one if there.
					if(this->m_AckQueue.size())
					{
						headers |= EQ_MSG_ACK;
						ackID = m_AckQueue[0].getAckID();
						m_AckQueue.remove(0);
					}
					else
						// Otherwise, clear the ACK header.
						headers &= ~EQ_MSG_ACK;
				}

				// if isMsg, then the whole thing is offset by 2 bytes...
				if(fragIndex == 0)
				{
					fragStart = txMsg->getData();

					fragSize = MAX_PACKET_SIZE - (isMsg ? 2 : 0);
				}
				else
				{
					fragStart=	txMsg->getData() + fragIndex * MAX_PACKET_SIZE - (isMsg ? 2 : 0);

					fragSize = txMsg->getDataSize() + (isMsg ? 2 : 0) - (fragIndex * MAX_PACKET_SIZE);
					if(fragSize > MAX_PACKET_SIZE)
						fragSize = MAX_PACKET_SIZE;
				}

				Send(new CEQTxPacket(
					headers | EQ_MSG_FRG | EQ_MSG_SEQ,
					m_NextTxPacketID++,
					ackID,
					m_NextTxSeqID++,
					m_NextTxFragID,									// frag id
					fragIndex,										// frag index
					numFrags,										// frag max
					(headers & EQ_MSG_MSG ? m_NextTxMsgID++ : 0),	// message id
					fragStart,										// data
					fragSize));										// size

				// EQ_MSG_MSG only goes out with the 1st packet in a fragmented
				// message. Also, if this is the 1st message, we sure don't wanna
				// resend SYN!
				headers &= ~(EQ_MSG_MSG | EQ_MSG_SYN);
			}
			m_NextTxFragID++;
		}
	}
	catch(CSocketException *e)
	{
		throw new CEQSocketException(e);
	}
}

// Send a CEQTxPacket to the client. Handles assuring the packet is ACKed if
// needed (if SEQ is set)
void CClientConnection::Send(CEQTxPacket *txPacket)
{
	CSingleLock localLock(&m_lockObj, TRUE);

	SendImpl(txPacket);

	if(txPacket->needsACK())
	{
		txPacket->setTimeSent();
		m_SendQueue.push_back(txPacket);
	}
	else
		delete txPacket;
}

// Sends the packet, psudo-blocks if send buffer is full. It would probably
// be better to have the game it's self lag in server frames than to allow
// the game to go on without user interaction if the internet connection
// is saturated. If this method blocks freqently, then the user limit is
// probably set too low for the connection speed.
void CClientConnection::SendImpl(CDatagramPacket *txPacket)
{
	txPacket->setInetAddr(this->m_addr);
	while(!m_serverSocket.m_socket->Send(*txPacket, true))
		Sleep(1);

	{
		CSingleLock localLock(&m_serverSocket.m_logFileSync, true);

		// log
		if(m_serverSocket.m_fDebugLog)
		{
			fprintf(m_serverSocket.m_fDebugLog, "Sending %s:\n", (const char*)CXTime().toString());
			hexdump(txPacket->getData().getData(), txPacket->getData().getDataSize(), 16, m_serverSocket.m_fDebugLog);
			fflush(m_serverSocket.m_fDebugLog);
		}
	}
}

// called when a new message is recieved.
void CClientConnection::MessageRecieved(CBinary *message)
{
	if(m_serverSocket.m_listener)
		m_serverSocket.m_listener->Recieve(this, message);
	else
		m_OrderedRecvQueue.push_back(message);
}

